from dateutil.parser import parse as date_parse
from typing import Any


def parse_value(val: str, dtype: str) -> Any:
    if val is None:
        return None
    
    val = str(val)
    if dtype == "str":
        return str(dtype)

    if dtype == "int":
        return int(val)

    if dtype == "date":
        return date_parse(val).strftime("%Y-%d-%m")

    if dtype == "datetime":
        return date_parse(val).strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]

    raise NotImplementedError(f"dtype {dtype} not implemented")



from yadet.engine.target.base import TargetFileSystemEngine, TargetEngine
from yadet.helpers.json import meta_handler

import gzip
import json
import os
from pathlib import Path
from typing import Dict, List


class FileSystemTargetEngine(TargetEngine):

    def __init__(self, base_directory: str, compression: str = None):
        super().__init__(vendor=TargetFileSystemEngine)
        self.base_directory = base_directory
        self.compression = compression
        # TODO: Partitions

    def batch_directory(self, batch_key: str) -> os.PathLike:
        return os.path.join(self.base_directory, batch_key)

    def write_batch(self, table_name: str, data: List[Dict], batch_key: str, start_indx: int):
        table_directory = os.path.join(self.batch_directory(batch_key), table_name)
        Path(table_directory).mkdir(parents=True, exist_ok=True)

        if self.compression is None:
            with open(os.path.join(table_directory, f"{batch_key}-{start_indx}.json"), "w") as f:
                json.dump(data, f)
                
        elif self.compression == "gzip":
            with gzip.open(os.path.join(table_directory, f"{batch_key}-{start_indx}.json.gz"), "w") as f:
                f.write(json.dumps(data).encode('utf-8'))
        
        else:
            raise NotImplementedError(f"Compression=`{self.compression}` not available")


    def write_meta(self, meta: Dict, batch_key: str):
        Path(self.batch_directory(batch_key)).mkdir(parents=True, exist_ok=True)

        with open(os.path.join(self.batch_directory(batch_key), "meta.json"), "w") as f:
            json.dump(meta, f, default=meta_handler, indent=4)


from yadet.config.table import TableConfig
from yadet.engine.source.base import SourceMssqlEngine, SourceEngine
from yadet.helpers.parser import parse_value
from yadet.helpers.sql import clean_sql

import pyodbc
from typing import Any, Dict


class MsSqlSourceEngine(SourceEngine):

    def __init__(self, connection_str: str, debug: bool = False):
        super().__init__(vendor=SourceMssqlEngine, debug=debug)
        self._connection_str: str = connection_str
        self._conn: pyodbc.Connection = None
        
        self.connect()

    def connect(self):
        self._conn: pyodbc.Connection = pyodbc.connect(self._connection_str)

    def fetch_one(self, sql: str) -> Any:
        with self.conn.cursor() as cursor:
            cursor.execute(sql)
            return cursor.fetchone()

    def fetch_all(self, sql: str) -> Any:
        with self.conn.cursor() as cursor:
            cursor.execute(sql)
            return cursor.fetchall()
            
    def where_clause(self, table_config: TableConfig, table_index: Dict) -> str:
        if not table_config.delta:
            return "" if table_config.filter_clause is None else f"WHERE {table_config.filter_clause}"
        
        where_clause = ""
        index_where_clause = ""
        if table_index:
            index_where_clause = " AND ".join([
                f"{field} > '{parse_value(val=table_index['columns'][field]['max'], dtype=dtype)}'" 
                for field, dtype in table_config.order_by_columns.items()
                if table_index.get("columns", {}).get(field, {}).get("max") is not None
            ])
        
        if table_config.filter_clause is not None:
            where_clause = f"WHERE {table_config.filter_clause}"
            if index_where_clause not in (None, ""):
                where_clause += f" AND {index_where_clause}"
        
        elif index_where_clause not in (None, ""):
            where_clause = f"WHERE {index_where_clause}"

        return where_clause

    def order_by_clause(self, table_config: TableConfig) -> str:
        order_by_field_clause = ", ".join([
            f"{field} ASC" for field in table_config.order_by_columns.keys()
        ])
        return f"ORDER BY {order_by_field_clause}"

    def join_clause(self, table_config: TableConfig) -> str:
        return "" if table_config.join_clause is None else table_config.join_clause

    def num_records_query(self, table_config: TableConfig, table_index: Dict) -> str:
        # Get Number of Records, Min, Max Value
        min_max_fields = ", ".join([
            f"{func}({field}) AS '{indx}__{func.lower()}'"
            for indx, field in enumerate(table_config.order_by_columns.keys())
            for func in ("MIN", "MAX")
        ])
        
        # Table Alias
        table_alias = "" if table_config.table_alias is None else f"AS {table_config.table_alias}"
        
        # SQL
        sql = f"""SELECT COUNT(*) AS num_records, {min_max_fields} FROM \
                    {table_config.table_name} {table_alias} {self.join_clause(table_config=table_config)} \
                    {self.where_clause(table_config=table_config, table_index=table_index)}"""
        return clean_sql(sql=sql)

    def extract_query(self, table_config: TableConfig, start_indx: int, table_index: Dict) -> str:
        # Table Alias
        table_alias = "" if table_config.table_alias is None else f"AS {table_config.table_alias}"
        
        base_sql = f"""
        SELECT {table_config.columns} 
        FROM {table_config.table_name} {table_alias} 
        {self.join_clause(table_config=table_config)}
        {self.where_clause(table_config=table_config, table_index=table_index)}
        {self.order_by_clause(table_config=table_config)}
        OFFSET {start_indx} ROWS
        FETCH NEXT {table_config.batch_size} ROWS ONLY
        """
        if table_config.flat:
            base_sql = f"WITH base_cte as ({base_sql}) SELECT * FROM base_cte"

        sql = f"{base_sql} FOR JSON AUTO;"
        return clean_sql(sql=sql)



from yadet.config.table import TableConfig
from yadet.engine.source.base import SourceMssqlEngine
from yadet.engine.source.mssql import MsSqlSourceEngine
from yadet.engine.target.base import TargetFileSystemEngine
from yadet.engine.target.file_system import FileSystemTargetEngine
from yadet.helpers.json import meta_handler
from yadet.objects.base import BaseObject

import json
import os
from pathlib import Path
from typing import Dict, List, Union


class ProjectConfig(BaseObject):

    def __init__(self, 
                 name: str,
                 source_engine: str = SourceMssqlEngine,
                 source_engine_params: Dict = {},
                 target_engine: str = TargetFileSystemEngine,
                 target_engine_params: Dict = {},
                 config_dir: Union[str, os.PathLike] = "config") -> None:
        self.name = name
        self.source_engine: str = source_engine
        self.source_engine_params: Dict = source_engine_params
        self.target_engine: str = target_engine
        self.target_engine_params: Dict = target_engine_params
        self.config_dir: Union[str, os.PathLike] = config_dir

        self._tables_config: List[TableConfig] = []
        self._project_table_index: Dict[str, Dict] = {}
        
        self.open()

    @property
    def project_directory(self) -> os.PathLike:
        return os.path.join(self.config_dir, self.name)
        
    @property
    def table_config_filename(self) -> os.PathLike:
        return os.path.join(self.project_directory, "project_config.json")

    @property
    def table_index_filename(self) -> os.PathLike:
        return os.path.join(
            self.target_engine_params["base_directory"], "project_table_index.json"
        )

    @property
    def tables_config(self) -> List[TableConfig]:
        return sorted(self._tables_config, key= lambda x: x.ordinal)

    @property
    def project_table_index(self) -> Dict[str, Dict]:
        return self._project_table_index

    @property
    def base_config(self) -> Dict:
        return {
            "name": self.name,
            "source_engine": self.source_engine,
            "target_engine": self.target_engine
        }
    
    def open(self):
        # Open Project Config
        if os.path.exists(self.table_config_filename):
            with open(self.table_config_filename, "r") as f:
                project_config = json.load(f)
                for config in project_config["items"]:
                    self.add_table_config(TableConfig(**config))
                # TODO: Add Engine
        else:
            Path(self.project_directory).mkdir(parents=True, exist_ok=True)
            with open(self.table_config_filename, "w") as f: 
                json.dump({**self.base_config, "items": []}, f)

        # Open Project Table Index
        if os.path.exists(self.table_index_filename):
            with open(self.table_index_filename, "r") as f:
                self._project_table_index = json.load(f)
        else:
            # Ensure Directory exists
            Path(
                os.path.dirname(self.table_index_filename)
            ).mkdir(parents=True, exist_ok=True)

            # Save Base Project Table Index
            with open(self.table_index_filename, "w") as f: 
                json.dump({}, f)

    def get_source_engine(self):
        if self.source_engine == SourceMssqlEngine:
            return MsSqlSourceEngine(**self.source_engine_params)
        raise NotImplementedError(f"Engine Type {self.source_engine} not active")
    
    def get_target_engine(self):
        if self.target_engine == TargetFileSystemEngine:
            return FileSystemTargetEngine(**self.target_engine_params)
        raise NotImplementedError(f"Engine Type {self.target_engine} not active")
        
    def save(self):
        Path(self.project_directory).mkdir(parents=True, exist_ok=True)
        with open(self.table_config_filename, "w") as f: 
            json.dump({
                **self.base_config, 
                "items": [table.data for table in self.tables_config]
            }, f, indent=4)

        with open(self.table_index_filename, "w") as f:
            json.dump(self.project_table_index, f, default=meta_handler, indent=4)
            
    def add_table_config(self, table_config: TableConfig):
        self._tables_config.append(table_config)

    def __str__(self) -> str:
        return str(self.name)


from yadet.config.table import TableConfig
from yadet.errors import ConfigException
from yadet.objects.base import BaseObject

from typing import Any

SourceMssqlEngine = "mssql"

SOURCE_ENGINE_OPTIONS = {SourceMssqlEngine, }


class SourceEngine(BaseObject):

    def __init__(self, vendor: str, debug: bool = False) -> None:
        self.vendor: str = vendor
        self.debug: bool = debug
        self._conn: Any = None
        self.validate()

    def validate(self):
        if self.vendor not in SOURCE_ENGINE_OPTIONS:
            raise ConfigException(
                f"Engine vendor=`{self.vendor}` not available, current options are `{', '.join(SOURCE_ENGINE_OPTIONS)}`"
            )
        if not hasattr(self, "connect"):
            raise ConfigException("Engine must have `connect` method")

    @property
    def conn(self) -> Any:
        return self._conn

    @property
    def active(self) -> bool:
        return self._conn is not None

    def where_clause(self, table_config: TableConfig) -> str:
        raise NotImplementedError()

    def order_by_clause(self, table_config: TableConfig) -> str:
        raise NotImplementedError()

    def num_records_query(self, table_config: TableConfig) -> str:
        raise NotImplementedError()

    def extract_query(self, table_config: TableConfig, start_indx: int) -> str:
        raise NotImplementedError()

    def __str__(self) -> str:
        return str(self.vendor)



from yadet.objects.base import BaseObject

from typing import Dict


class TableConfig(BaseObject):
    
    def __init__(self, 
                 ordinal: int, 
                 table_name: str,
                 delta: bool,
                 order_by_columns: Dict[str, str],
                 table_alias: str = None,
                 output_table_name: str = None,
                 columns: str = "*",
                 filter_clause: str = None,
                 join_clause: str = None,
                 flat: bool = True,
                 batch_size: int = 5000,
                 min_wait: int = None,
                 active: bool = True,
                 desc: str = None):
        self.ordinal: int = ordinal
        self.table_name: str = table_name
        self.delta: bool = delta
        self.order_by_columns: Dict[str, str] = order_by_columns
        self.table_alias: str = table_alias
        self.output_table_name: str = output_table_name
        self.columns: str = columns
        self.filter_clause: str = filter_clause
        self.join_clause: str = join_clause
        self.flat: bool = flat
        self.batch_size: int = batch_size
        self.min_wait: int = min_wait
        self.active: bool = active
        self.desc: str = desc

        self.validate()

    def validate(self):
        # TODO: Validate structure of elements
        # TODO: Validate Order by Columns
        ...

    @property
    def get_table_alias(self) -> str:
        return self.table_alias or self.table_name

    @property
    def get_output_table_name(self) -> str:
        return self.output_table_name or self.get_table_alias

    @property
    def data(self) -> Dict:
        return {
            "ordinal": self.ordinal,
            "table_name": self.table_name,
            "delta": self.delta,
            "order_by_columns": self.order_by_columns,
            "table_alias": self.table_alias,
            "output_table_name": self.output_table_name,
            "columns": self.columns,
            "filter_clause": self.filter_clause,
            "join_clause": self.join_clause,
            "flat": self.flat,
            "batch_size": self.batch_size,
            "min_wait": self.min_wait,
            "active": self.active,
            "desc": self.desc,
        }
        
    def __str__(self) -> str:
        return f"{str(self.table_alias)}: {self.ordinal}"


from yadet.config.table import TableConfig
from yadet.interface.source import SourceInterface
from yadet.interface.target import TargetInterface
from yadet.objects.base import BaseObject

from datetime import datetime
from typing import Dict


class ExtractTransferInterface(BaseObject):

    def __init__(self,
                 source: SourceInterface,
                 target: TargetInterface,
                 table_config: TableConfig) -> None:
        self.source: SourceInterface = source
        self.target: TargetInterface = target
        self.table_config: TableConfig = table_config
        self.meta: Dict = {}

    def info(self, msg: str):
        print(f"[{self.table_config.get_table_alias}] {msg}")

    def start(self):
        self.meta["table"] = self.table_config.get_table_alias
        self.meta["start_datetime"] = datetime.now()

    def end(self):
        self.meta["end_datetime"] = datetime.now()
        self.meta["duration_seconds"] = (
            self.meta["end_datetime"] - self.meta["start_datetime"]
        ).total_seconds()

    def extract_transfer(self, batch_key: str) -> Dict:
        self.meta = self.source.fetch_num_records()

        # Start Tracking
        self.start()

        self.info(f"Collecting {self.meta['records']:,.0f} records")

        start_indx = 0
        while start_indx < self.meta["records"]:

            # Extract Data
            data = self.source.fetch_extract_batch(start_indx=start_indx)

            # Write Batch Out
            self.target.write_batch(
                data=data, batch_key=batch_key, start_indx=start_indx
            )

            # Increment Index
            start_indx += self.table_config.batch_size
            self.info(f"Competed batch start_indx={start_indx:,.0f}")

        # End Tracking
        self.end()

        return self.meta

    def __str__(self) -> str:
        return f"{str(self.source)} -> {str(self.target)}"



from yadet.config.project import ProjectConfig
from yadet.interface.extract_transfer import ExtractTransferInterface
from yadet.interface.source import SourceInterface
from yadet.interface.target import TargetInterface
from yadet.objects.base import BaseObject

from datetime import datetime, timedelta
from dateutil.parser import parse as date_parse
from typing import Any, Dict, Optional


class ProjectBatchRun(BaseObject):

    def __init__(self,
                 project: ProjectConfig,
                 batch_id: Optional[Any] = None,
                 debug: bool = False):
        self.project: ProjectConfig = project
        self.batch_id: Optional[Any] = batch_id
        self.batch_key: str = datetime.today().strftime("%y%m%d-%H%M")
        self.debug: bool = debug
        self._meta: Dict = {}

    def start(self):
        self._meta["batch_id"] = self.batch_id
        self._meta["batch_key"] = self.batch_key
        self._meta["start_datetime"] = datetime.now()
        self._meta["table_meta"] = []

    def end(self):
        self._meta["end_datetime"] = datetime.now()
        self._meta["duration_seconds"] = (
            self._meta["end_datetime"] - self._meta["start_datetime"]
        ).total_seconds()

        # Write Meta for Batch
        self.target_engine.write_meta(meta=self.meta, batch_key=self.batch_key)

        # Update Table Index for Project
        if not self.debug:
            for table_meta in self.meta["table_meta"]:
                if table_meta["records"] > 0:
                    self.project.project_table_index.update({
                        table_meta["table"]: {
                            "last_runtime": self._meta["start_datetime"],
                            "columns": table_meta["columns"]
                        }
                    })

        self.project.save()

    def add_table_meta(self, table_meta: Dict):
        self._meta["table_meta"].append(table_meta)

    def run_batch(self):
        self.start()
        self.source_engine = self.project.get_source_engine()
        self.target_engine = self.project.get_target_engine()

        for table_config in self.project.tables_config:
            # Skip if Table Config not Active
            if not table_config.active:
                # TODO: Add Meta for Skipped Table
                print(
                    f'[{table_config.get_table_alias}] Skipping inactive table'
                )
                continue

            # Source Table Index
            table_index = self.project.project_table_index.get(
                table_config.get_table_alias, {}
            )

            # Skip if Min Wait is Active
            if table_config.min_wait is not None:
                last_runtime = table_index.get("last_runtime")
                if last_runtime is not None:
                    last_runtime = date_parse(last_runtime)
                    next_runtime = last_runtime + timedelta(seconds=table_config.min_wait * 60)
                    if next_runtime > datetime.now():
                        print(f'[{table_config.get_table_alias}] Skipping table with min wait: {str(next_runtime)}')
                        continue
                    
            source = SourceInterface(
                engine=self.source_engine, table_config=table_config, 
                table_index=table_index
            )
            target = TargetInterface(
                engine=self.target_engine, table_config=table_config
            )
            extact_transfer = ExtractTransferInterface(
                source=source, target=target, table_config=table_config
            )

            # Run Transfer
            table_meta = extact_transfer.extract_transfer(
                batch_key=self.batch_key
            )
            self.add_table_meta(table_meta=table_meta)

        self.end()

    @property
    def meta(self) -> Dict:
        return self._meta



from datetime import datetime, date


def meta_handler(obj):
    """Datetime Handlers (TODO: Have Converters for Various SQL Variants)

    Parameters
    ----------
    obj : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    if isinstance(obj, (datetime, date)):
        return obj.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]

    if isinstance(obj, None):
        return None



# -*- coding: utf-8 -*-
from setuptools import setup


install_requires = ["pyodbc"]

setup(
    name="yadet",
    version="1.3.0",
    author="Robert Goss",
    author_email="robertgoss07@gmail.com",
    packages=[
        "yadet", "yadet.batch", "yadet.config", "yadet.engine", "yadet.engine.source", 
        "yadet.engine.target", "yadet.errors", "yadet.helpers", "yadet.interface", 
        "yadet.objects"
    ],
    license="MIT",
    extras = {"extras": install_requires},
    url="https://github.com/robbitt07/yadet",
    install_requires=install_requires,
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    description="A simple and effective Python data extraction tool",
    long_description=open("README.md").read(),
    zip_safe=True,
)


from yadet.config.table import TableConfig
from yadet.engine.target.base import TargetEngine
from yadet.objects.base import BaseObject

from typing import Dict, List


class TargetInterface(BaseObject):

    def __init__(self, engine: TargetEngine, table_config: TableConfig) -> None:
        self.engine: TargetEngine = engine
        self.table_config: TableConfig = table_config

    def write_batch(self, data: List[Dict], batch_key: str, start_indx: int):
        return self.engine.write_batch(
            table_name=self.table_config.get_output_table_name, data=data, batch_key=batch_key,
            start_indx=start_indx
        )

    def write_meta(self, meta: Dict, batch_key: str):
        return self.engine.write_meta(meta=meta, batch_key=batch_key)

    def __str__(self) -> str:
        return f"{str(self.engine)}"



from typing import Dict, List

from yadet.objects.base import BaseObject

TargetFileSystemEngine = "file_system"

TARGET_ENGINE_OPTIONS = {TargetFileSystemEngine, }


class TargetEngine(BaseObject):

    def __init__(self, vendor: str) -> None:
        self.vendor: str = vendor

    def write_batch(self, data: List[Dict], batch_key: str, start_indx: int):
        raise NotImplementedError()

    def write_meta(self, meta: Dict, batch_key: str):
        raise NotImplementedError()

    def __str__(self):
        return str(self.vendor)



from yadet.config.table import TableConfig
from yadet.engine.source.base import SourceEngine
from yadet.helpers.base import chunks
from yadet.helpers.parser import parse_value
from yadet.objects.base import BaseObject

import json
from typing import Dict, List


class SourceInterface(BaseObject):

    def __init__(self, engine: SourceEngine, table_config: TableConfig, table_index: Dict) -> None:
        self.engine: SourceEngine = engine
        self.table_config: TableConfig = table_config
        self.table_index: Dict = table_index

    def info(self, msg: str):
        print(f"[{self.table_config.get_table_alias}] {msg}")

    def num_records_query(self) -> str:
        query = self.engine.num_records_query(
            table_config=self.table_config, table_index=self.table_index
        )
        if self.engine.debug:
            self.info(query)
        return query

    def extract_query(self, start_indx: int) -> str:
        query = self.engine.extract_query(
            table_config=self.table_config, start_indx=start_indx, table_index=self.table_index
        )
        if self.engine.debug:
            self.info(query)
        return query

    def fetch_num_records(self) -> Dict:
        records, *order_range = self.engine.fetch_one(self.num_records_query())
        return {
            "records": records,
            "columns": {
                col: {
                    "min": parse_value(val=min_max[0], dtype=dtype),
                    "max": parse_value(val=min_max[1], dtype=dtype)
                }
                for (col, dtype), min_max
                in zip(self.table_config.order_by_columns.items(), list(chunks(order_range)))
            }}

    def fetch_extract_batch(self, start_indx: int) -> List[Dict]:
        rows = self.engine.fetch_all(self.extract_query(start_indx=start_indx))
        if len(rows) > 0:
            return json.loads(''.join([row[0] for row in rows]))
        return []

    def __str__(self) -> str:
        return f"{str(self.engine)}"
